<?php

namespace Drupal\template_generator\Form;

use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Entity\ContentEntityType;

/**
 * Class for form generate template settings.
 */
class GenerateTemplateForm extends FormBase {

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'template_generator_form';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $datas = \Drupal::state()->get('template_generator') ?: [];
    $form['#tree'] = TRUE;
    $entities_informations = $this->getEntitiesInformations();
    $form['entities_enabled'] = [
      '#title' => '<h2>' . t('Template entities to generate') . '</h2>',
      '#type' => 'checkboxes',
      '#options' => array_column($entities_informations, 'name', 'id'),
      '#group' => 'advanced',
      '#default_value' => isset($datas['entities_enabled']) ? $datas['entities_enabled'] : [],
    ];
    $form['entities'] = [
      '#type' => 'item',
      '#title' => '<h2>' . t('Parameter by entity') . '</h2>',
    ];

    // List of table of each bundle of each entity.
    foreach ($entities_informations as $entity_id => $entity_infos) {
      $form['entities'][$entity_id . '_fields'] = [
        '#type' => 'item',
        '#title' => "<h3>" . $entity_infos['name'] . "</h3>",
        '#states' => [
          'visible' => [
            ':input[name="entities_enabled' . '[' . $entity_id . ']"]' => ['checked' => TRUE],
          ],
        ],
      ];
      $form['entities'][$entity_id . '_fields']['mode'] = [
        '#type' => 'select',
        '#options' => [
          'none' => t('No sorting'),
          'view_mode' => t('Order by View Mode'),
          'bundle' => t('Order by Bundle'),
        ],
        '#default_value' => isset($datas['entities'][$entity_id . '_fields']['mode']) ? $datas['entities'][$entity_id . '_fields']['mode'] : [],
      ];

      // Fieldset to ignore certain view mode or bundle.
      $form['entities'][$entity_id . '_fields']['ignore'] = [
        '#title' => t("Ignore specific bundle or view mode"),
        '#type' => 'details',
      ];
      $form['entities'][$entity_id . '_fields']['ignore']['bundle'] = [
        '#title' => t("Disabled for the following bundles :"),
        '#type' => 'checkboxes',
        '#options' => $entity_infos['bundles_list'],
        '#default_value' => isset($datas['entities'][$entity_id . '_fields']['ignore']['bundle']) ? $datas['entities'][$entity_id . '_fields']['ignore']['bundle'] : [],
      ];
      $form['entities'][$entity_id . '_fields']['ignore']['viewmode'] = [
        '#title' => t("Disabled for the following view modes :"),
        '#type' => 'checkboxes',
        '#options' => $entity_infos['viewmodes_list'],
        '#default_value' => isset($datas['entities'][$entity_id . '_fields']['ignore']['viewmode']) ? $datas['entities'][$entity_id . '_fields']['ignore']['viewmode'] : [],
      ];
    }
    $form['actions']['submit'] = [
      '#type' => 'submit',
      '#value' => $this->t('Save and Generate'),
      '#button_type' => 'primary',
    ];
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $datas = [
      'entities_enabled' => $form_state->getValue('entities_enabled'),
      'entities' => $form_state->getValue('entities'),
    ];
    \Drupal::state()->set('template_generator', $datas);
    $this->generate($datas);
    \Drupal::messenger()->addMessage('Templates generated');
  }

  /**
   * Get information of entities like bundles or view mode.
   *
   * @return array
   *   Entities informations.
   */
  protected function getEntitiesInformations() {
    $data = [];

    // Get definitions of entities.
    $entity = \Drupal::service('entity_display.repository');
    $entity_type_definations = \Drupal::entityTypeManager()->getDefinitions();

    // Get all content entities only.
    foreach ($entity_type_definations as $definition) {
      if ($definition instanceof ContentEntityType) {

        // Get all bundle of this entity.
        $bundles_infos = \Drupal::service('entity_type.bundle.info')
          ->getBundleInfo($definition->id());
        $bundles = array_keys($bundles_infos);

        // Retrieval of all activated view mode for a specific bundle.
        foreach ($bundles as $bundle) {
          $vm = $entity->getViewModeOptionsByBundle($definition->id(), $bundle);
          if (count($vm) > 0) {
            $data[$definition->id()]['bundles'][$bundle] = array_keys($vm);
          }
        }

        // Set information for this entity.
        if (isset($data[$definition->id()]['bundles'])) {
          $label = $definition->getLabel();
          $data[$definition->id()]['name'] = render($label);
          $data[$definition->id()]['id'] = $definition->id();

          // Set all avalaible bundle for this entity.
          $data[$definition->id()]['bundles_list'] = [];
          foreach ($bundles_infos as $name => $info) {
            $data[$definition->id()]['bundles_list'][$name] = (string) $info['label'];
          }

          // Set all avalaible view mode for this entity.
          $data[$definition->id()]['viewmodes_list'] = [];
          $entitydisplay = \Drupal::service('entity_display.repository');
          $entitydisplay = $entitydisplay->getViewModes($definition->id());

          foreach ($entitydisplay as $name => $info) {
            $data[$definition->id()]['viewmodes_list'][$name] = (string) $info['label'];
          }
        }
      }
    }
    return $data;
  }

  /**
   * Prepare Generate Templates.
   *
   * @param array $values
   *   Values from form generate.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  protected function generate(array $values) {
    $entities = $this->getEntitiesInformations();
    $parent_path = \Drupal::root() . '/' . drupal_get_path('theme', \Drupal::config('system.theme')
      ->get('default'));

    // Generate templates for all desired entities.
    foreach ($entities as $entity) {
      if ($values['entities_enabled'][$entity['id']] !== 0) {
        $bundles_ignores = array_flip($values['entities'][$entity['id'] . '_fields']['ignore']['bundle']);
        $vm_ignores = array_flip($values['entities'][$entity['id'] . '_fields']['ignore']['viewmode']);
        foreach ($entity['bundles'] as $bundle_id => $bundle) {
          foreach ($bundle as $view_mode) {
            // Do not put view mode in the template name when it is the default.
            if (!isset($bundles_ignores[$bundle_id]) && !isset($vm_ignores[$view_mode])) {

              $html = NULL;
              $view_mode_tpl = str_replace('_', '-', $view_mode);
              if ($view_mode == 'default') {
                $end = $bundle_id;
              }
              else {
                $end = $bundle_id . '--' . $view_mode_tpl;
              }
              $path = $parent_path . '/templates/' . $entity['id'];

              // Change of folder depending on the option chosen.
              switch ($values['entities'][$entity['id'] . '_fields']['mode']) {
                case 'view_mode':
                  $path .= '/' . $view_mode_tpl;
                  break;

                case 'bundle':
                  $path .= '/' . $bundle_id;
                  break;

                default:
                  break;
              }
              $filename = $entity['id'] . '--' . $end . '.html.twig';
              $full_path = $path . '/' . $filename;

              // Create file only is not exist.
              if (!file_exists($full_path)) {
                if (!file_exists($path)) {
                  \Drupal::service('file_system')->mkdir($path, NULL, TRUE);
                }
              }

              // If file already exist, only change parameters.
              else {
                $html = file_get_contents($full_path);
                $val = strpos($html, "{#");
                if ($val === 0) {
                  $end = strpos($html, "#}");
                  if ($end !== FALSE) {
                    $html = trim(substr($html, $end + 2));
                  }
                }
              }
              file_put_contents($full_path, $this->template($entity['id'], $bundle_id, $view_mode, $html));
            }
          }
        }
      }
    }
  }

  /**
   * Writing the template file.
   *
   * @param string $entity_type
   *   The name of the entity.
   * @param string $bundle
   *   The name of the bundle.
   * @param string $view_mode
   *   The name of the view mode.
   * @param string $html
   *   HTML of content.
   *
   * @return string
   *   Generated template.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  protected function template($entity_type, $bundle, $view_mode, $html = NULL) {

    // Get information of this view mode of bundle.
    $type = \Drupal::entityTypeManager()
      ->getStorage('entity_view_display')
      ->load($entity_type . '.' . $bundle . '.' . $view_mode);
    $type = $type->toArray();

    // Get all availables fields.
    $fields = '';
    $field_list = [];
    foreach ($type['content'] as $name => $content) {
      $field = \Drupal::entityTypeManager()
        ->getStorage('field_storage_config')
        ->load($entity_type . '.' . $name);
      if (!$field) {
        continue;
      }
      $field = $field->toArray();

      // Prepare Header.
      $fields .= '* - ' . $name . PHP_EOL;
      $fields .= '* | type: ' . $field['type'] . PHP_EOL;
      $fields .= '* | cardinality: ' . $field['cardinality'] . PHP_EOL;
      $field_list[] .= '{{ content.' . $name . ' }}';
    }

    // Return Existing Html or list of available fields.
    return '{#
/******
* @file
* Default theme implementation to display a ' . $entity_type . '
*
* Availables fields :
'
      . $fields .
      '******/
#}

' .
      ($html ?: implode(PHP_EOL, $field_list)) . "\n";
  }

}
